// ------------------------------------------------------------------------
// Convex Hull -- Tecgraf/PUC-Rio/IMPA
// www.tecgraf.puc-rio.br/~lduarte
//
// $Source:  $
// $Revision: 1.0 $
// $Date: 2008/09/06 19:21:00 $
//
// $Author: lduarte $
// ------------------------------------------------------------------------

#ifdef WIN32
#include <windows.h>
#endif

#include <gl/gl.h>
#include <gl/glu.h>
#include <float.h>
#include <algorithm>
#include <vr/timer.h>

#include "convex_hull.h"
#include "util.h"

extern void displayScreenMessage( const char* s, float x, float y );

static vr::Timer timer;

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::reset(std::vector<cCoord>& pts)
{
 minX.x = DBL_MAX;
 minX.y = DBL_MAX;

 maxX.x = -DBL_MAX;
 maxX.y = -DBL_MAX;

 minY.x = DBL_MAX;
 minY.y = DBL_MAX;

 maxY.x = -DBL_MAX;
 maxY.y = -DBL_MAX;

 points.clear();
 result.clear();
 sorted_points.clear();

 num_points = (int)pts.size();

 points.resize(num_points);
 sorted_points.resize(num_points);
 result.resize(num_points);

 has_result = false;
 has_elimination = false;

 for(int i = 0; i < num_points; i++)
 {
  points[i] = pts[i];
  sorted_points[i] = pts[i];
  
  minX = lexoMinX(minX, points[i]);
  maxX = lexoMaxX(maxX, points[i]);
  minY = lexoMinY(minY, points[i]);
  maxY = lexoMaxY(maxY, points[i]);
 }
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::interior_point_elimination(void)
{
 minX.x = DBL_MAX;
 minX.y = DBL_MAX;

 maxX.x = -DBL_MAX;
 maxX.y = -DBL_MAX;

 minY.x = DBL_MAX;
 minY.y = DBL_MAX;

 maxY.x = -DBL_MAX;
 maxY.y = -DBL_MAX;

 // finding bbox
 for(int i = 0; i < num_points; i++)
 {
  minX = lexoMinX(minX, points[i]);
  maxX = lexoMaxX(maxX, points[i]);
  minY = lexoMinY(minY, points[i]);
  maxY = lexoMaxY(maxY, points[i]);
 }

 int it = 0;
 cCoord u, v;
 bool r1, r2, r3, r4;
 for(int i = 0; i < num_points; i++)
 {
  u = maxX - minY;
  v = points[i] - minY;
  r1 = mostUpRight(u, v);

  u = maxY - maxX;
  v = points[i] - maxX;
  r2 = mostUpRight(u, v);

  u = minX - maxY;
  v = points[i] - maxY;
  r3 = mostUpRight(u, v);

  u = minY - minX;
  v = points[i] - minX;
  r4 = mostUpRight(u, v);

  if(r1 || r2 || r3 || r4)
   points[it++] = points[i];
 }
 num_points = it;
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::jarvis(void)
{
 timer.restart();

 if(has_elimination)
  this->interior_point_elimination();

 cCoord p1, p2, v1, v2;

 // minimum lexographical point
 p1 = minY;
 p2 = minY;

 // initial horizonal line
 p1.x += p1.x > 0.0 ? (-p1.x/2.0) : (p1.x/2.0);

 int it = 0;

 // end of the convex hull
 do {
  v1 = p1 - p2;
  p1 = p2;

  // p1 already belongs to the convex hull
  result[it++] = p1;

  // testing each point
  for(int i = 0; i < num_points; i++)
  {
   // distinct points
   if(points[i].id != p1.id)
   {
    // next convex hull candidate
    v2 = points[i] - p1;

    // if the point is up right of v1
    if(this->mostUpRight(v1, v2))
    {
     // the current most up right
     p2 = points[i];
     v1 = p2 - p1;
    }
   }
  }
 } while(p2 != minY);

 num_points_result = it;

 has_result = true;

 double jarvis_time = timer.elapsed();
 printf("Jarvis Algorithm : %.4e ms  %d\n", jarvis_time * 1000, num_points_result);
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
bool cConvexHull::mostUpRight(const cCoord& u, const cCoord &v)
{
 double det = detCrossProd(u, v);

 if(det > 0.0)
  return false;
 else if(det < 0.0)
  return true;
 else
 {
  if(v.x > u.x)
   return true;
  else if(v.x < u.x)
   return false;
  else if(u.x == v.x)
  {
   if(v.y > u.y)
    return true;
   else if(v.y < u.y)
    return false;
  }
 }
 return true;
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::scanning(void)
{
 if(num_points < 2)
  return;

 timer.restart();

 if(has_elimination)
  this->interior_point_elimination();

 sorted_points.resize(num_points);
 for(int i = 0; i < num_points; i++)
  sorted_points[i] = points[i];

 // lexographical sort
 this->merge_sort();

 int i, it = 0;
 cCoord v1, v2;

 result[it++] = sorted_points[0];
 result[it++] = sorted_points[1];

 for(i = 2; i < num_points; i++)
 {
  result[it++] = sorted_points[i];
  while(it > 2 && !(detCrossProd(result[it-3] - result[it-2], result[it-1] - result[it-2]) < 0.0))
  {
   result[it-2] = result[it-1];
   it--;
  }
 }

 vector<cCoord> tmp_result(num_points);

 int itt = 0;
 tmp_result[itt++] = sorted_points[num_points-1];
 tmp_result[itt++] = sorted_points[num_points-2];

 for(int j = num_points-3; j >= 0; --j)
 {
  tmp_result[itt++] = sorted_points[j];
  while(itt > 2 && !(detCrossProd(tmp_result[itt-3] - tmp_result[itt-2], tmp_result[itt-1] - tmp_result[itt-2]) < 0.0))
  {
   tmp_result[itt-2] = tmp_result[itt-1];
   itt--;
  }
 }

 for(int k = 1; k < itt-1; k++)
 {
  if(it < num_points)
   result[it++] = tmp_result[k];
 }

 num_points_result = it;

 has_result = true;

 double scan_time = timer.elapsed();
 printf("Scanning Algorithm : %.4e ms  %d\n", scan_time * 1000, num_points_result);
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::merge_sort(void)
{
 if(num_points < 2)
  return;

 sort(&sorted_points);
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::sort(vector<cCoord>* v)
{
 int size = (int)v->size();

 // if 1 element need no sort
 if(size < 2)
  return;

 int i, j, half = size / 2;
 vector<cCoord> v1(half), v2(size - half);

 // dividing into two halves
 for(i = 0; i < half; i++)
  v1[i] = (*v)[i];

 for(j = 0, i = half; i < size; j++, i++)
  v2[j] = (*v)[i];

 // sorting the halves 
 sort(&v1);
 sort(&v2);

 // merging the result
 merge(&v1, &v2, v);
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::merge(vector<cCoord>* v1, vector<cCoord>* v2, vector<cCoord>* res)
{
 int size1 = v1->size();
 int size2 = v2->size();

 int i = 0, it1 = 0, it2 = 0;

 // stop when the two vecs finish
 while(it1 < size1 || it2 < size2)
 {
  if(it1 >= size1)
  {
   (*res)[i++] = (*v2)[it2++];
   continue;
  }
  else if(it2 >= size2)
  {
   (*res)[i++] = (*v1)[it1++];
   continue;
  }

  // if x is equal, check y
  if(fabs((*v1)[it1].x - (*v2)[it2].x) < TOL_ZERO)
  {
   // if y is equal too, put anyone and increment pointer
   if(fabs((*v1)[it1].y - (*v2)[it2].y) < TOL_ZERO)
    (*res)[i++] = (*v1)[it1++];
   // put the less one and increment pointer
   else if((*v1)[it1].y < (*v2)[it2].y)
    (*res)[i++] = (*v1)[it1++];
   else
    (*res)[i++] = (*v2)[it2++];
  }
  // put the less one and increment pointer
  else
   if((*v1)[it1].x < (*v2)[it2].x)
    (*res)[i++] = (*v1)[it1++];
   else
    (*res)[i++] = (*v2)[it2++];
 }
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
cCoord cConvexHull::lexoMinX(cCoord p1, cCoord p2)
{
 if(p1.x == p2.x)
 {
  if(p1.y == p2.y)
   return p1;
  else if(p1.y < p2.y)
   return p1;
  else
   return p2;
 }
 else if(p1.x < p2.x)
  return p1;
 else
  return p2;
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
cCoord cConvexHull::lexoMaxX(cCoord p1, cCoord p2)
{
 if(p1.x == p2.x)
 {
  if(p1.y == p2.y)
   return p1;
  else if(p1.y > p2.y)
   return p1;
  else
   return p2;
 }
 else if(p1.x > p2.x)
  return p1;
 else
  return p2;
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
cCoord cConvexHull::lexoMinY(cCoord p1, cCoord p2)
{
 if(p1.y == p2.y)
 {
  if(p1.x == p2.x)
   return p1;
  else if(p1.x < p2.x)
   return p1;
  else
   return p2;
 }
 else if(p1.y < p2.y)
  return p1;
 else
  return p2;
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
cCoord cConvexHull::lexoMaxY(cCoord p1, cCoord p2)
{
 if(p1.y == p2.y)
 {
  if(p1.x == p2.x)
   return p1;
  else if(p1.x > p2.x)
   return p1;
  else
   return p2;
 }
 else if(p1.y > p2.y)
  return p1;
 else
  return p2;
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::draw(void)
{
 if(!has_result)
  return;

 glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LINE_BIT | GL_POINT_BIT | GL_COLOR_BUFFER_BIT);

 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

 //glEnable(GL_POINT_SMOOTH);
 //glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);

 glLineWidth(2.0f);
 glPointSize(5.0);

 glColor4f(1.0f, 0.0f, 0.0f, 1.0f);

 glBegin(GL_POINTS);
 for(int i = 0; i < (int)points.size(); i++)
  glVertex3d(points[i].x, points[i].y, 0.0);
 glEnd();

 char s[5];
 for(int i = 0; i < num_points_result; i++)
 {
  glBegin(GL_POINTS);
  glVertex3d(result[i].x, result[i].y, 0.0);
  glEnd();

  if(num_points_result < 100)
  {
   sprintf(s, "%c", result[i].id);
   displayScreenMessage((const char*)s, (float)result[i].x + 0.1f, (float)result[i].y + 0.1f);
  }
 }

 glColor4f(0.0f, 0.0f, 1.0f, 0.5f);

 glBegin(GL_POLYGON);
 for(int i = 0; i < num_points_result; i++)
  glVertex3d(result[i].x, result[i].y, 0.0);
 glVertex3d(result[0].x, result[0].y, 0.0);
 glEnd();

 glPopAttrib();
}

// ------------------------------------------------------------------------
// ------------------------------------------------------------------------
void cConvexHull::draw_elimination(void)
{
 glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT | GL_LINE_BIT | GL_POINT_BIT | GL_COLOR_BUFFER_BIT);

 glEnable(GL_BLEND);
 glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

 glLineWidth(2.0f);
 glPointSize(5.0);

 glColor4f(1.0f, 0.0f, 0.0f, 1.0f);

 glBegin(GL_POINTS);
 for(int i = 0; i < num_points; i++)
  glVertex3d(points[i].x, points[i].y, 0.0);
 glEnd();

 glColor4f(0.0f, 0.0f, 1.0f, 1.0f);

 glBegin(GL_LINE_STRIP);
 glVertex2d(minY.x, minY.y);
 glVertex2d(maxX.x, maxX.y);
 glVertex2d(maxY.x, maxY.y);
 glVertex2d(minX.x, minX.y);
 glVertex2d(minY.x, minY.y);
 glEnd();

 displayScreenMessage("minY", (float)minY.x + 0.1f, (float)minY.y + 0.1f);
 displayScreenMessage("maxX", (float)maxX.x + 0.1f, (float)maxX.y + 0.1f);
 displayScreenMessage("maxY", (float)maxY.x + 0.1f, (float)maxY.y + 0.1f); 
 displayScreenMessage("minX", (float)minX.x + 0.1f, (float)minX.y + 0.1f);

 glPopAttrib();
}
