/**
*	Exemplo grid geometry shader
*/

#include <stdio.h>
#include <time.h>

#include <vr/random.h>
#include <vr/timer.h>
#include <vr/string.h>
#include <vr/mat4.h>
#include <vr/vec2.h>
#include <vr/stl_utils.h>

#include <gl/glew.h>
#include <gl/glut.h>
#include <gl/wglew.h>

#include "domain.h"
#include "convex_hull.h"
#include "delaunay.h"

/************************************************************************/
/* Global variables                                                     */
/************************************************************************/

// Global objects
static unsigned int nPoints = 1000; 
static double minValue = 0.0;
static double maxValue = 20.0;
static cDomain* domain = NULL;
static cConvexHull* convex_hull = NULL;
static cDelaunay* delaunay = NULL;

// Camera manipulation
static double s_walkSpeed = 0.5;
static double s_lookSpeed = 0.005;

static vr::mat4f s_viewMatrix;
static vr::mat4f s_projMatrix;

static vr::vec2f s_prevMousePos = vr::vec2f( 0.0, 0.0 );
static bool s_mouseDragActive = false;

// Measuring fps
static int s_frameCounter = 0;
static vr::Timer s_fpsTimer;
static vr::String s_fpsMsg;

// Miscellaneous
static bool s_wireframeActive = false;
static int s_screenWidth  = 640;
static int s_screenHeight = 480;

int num_operations;

/************************************************************************/
/* Main functions                                                       */
/************************************************************************/

//------------------------------------------------------------------------
//------------------------------------------------------------------------
static void defaultScene(void)
{
 if(!domain)
  return;

 cCoord p;

 p.id = A;
 p.x = 3.0;
 p.y = 9.0;
 domain->addPoint(p);

 p.id++;
 p.x = 11.0;
 p.y = 1.0;
 domain->addPoint(p);

 
 
 
 
 //p.id++;
 //p.x = 1.0;
 //p.y = 1.0;
 //domain->addPoint(p);

 //p.id++;
 //p.x = 5.0;
 //p.y = 5.0;
 //domain->addPoint(p);

 //p.id++;
 //p.x = 8.0;
 //p.y = 5.0;
 //domain->addPoint(p);

 //p.id++;
 //p.x = 3.0;
 //p.y = 4.0;
 //domain->addPoint(p);
 
 
 
 
 p.id++;
 p.x = 6.0;
 p.y = 8.0;
 domain->addPoint(p);

 p.id++;
 p.x = 4.0;
 p.y = 3.0;
 domain->addPoint(p);

 p.id++;
 p.x = 5.0;
 p.y = 15.0;
 domain->addPoint(p);

 p.id++;
 p.x = 8.0;
 p.y = 11.0;
 domain->addPoint(p);

 p.id++;
 p.x = 1.0;
 p.y = 6.0;
 domain->addPoint(p);

 p.id++;
 p.x = 7.0;
 p.y = 4.0;
 domain->addPoint(p);

 p.id++;
 p.x = 9.0;
 p.y = 7.0;
 domain->addPoint(p);

 p.id++;
 p.x = 14.0;
 p.y = 5.0;
 domain->addPoint(p);

 p.id++;
 p.x = 10.0;
 p.y = 13.0;
 domain->addPoint(p);

 p.id++;
 p.x = 17.0;
 p.y = 14.0;
 domain->addPoint(p);

 p.id++;
 p.x = 15.0;
 p.y = 2.0;
 domain->addPoint(p);

 p.id++;
 p.x = 13.0;
 p.y = 16.0;
 domain->addPoint(p);

 p.id++;
 p.x = 3.0;
 p.y = 12.0;
 domain->addPoint(p);

 p.id++;
 p.x = 12.0;
 p.y = 10.0;
 domain->addPoint(p);

 p.id++;
 p.x = 16.0;
 p.y = 8.0;
 domain->addPoint(p);

 domain->set_format("%c");
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
static void createScene()
{
 domain = new cDomain();

 defaultScene();

 convex_hull = new cConvexHull(domain->get_points());
 delaunay = new cDelaunay(domain->get_points());
 //delaunay->divide_and_conquer();
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Draw scene
static void drawScene()
{
 if( domain == NULL )
  return;

 /*if(convex_hull->hasResult())
  convex_hull->draw();
 else if(convex_hull->hasElimination())
  convex_hull->draw_elimination();
 else
  domain->draw();*/

 if(delaunay->hasResult())
  delaunay->draw();
 
 if(domain->hasDomain())
  domain->draw();
}

/************************************************************************/
/* Utility functions                                                    */
/************************************************************************/

//------------------------------------------------------------------------
//------------------------------------------------------------------------
static void resetCamera()
{
 s_viewMatrix.makeLookAt( vr::vec3f( 0.0, 0.0, 20.0 ), vr::vec3f( 0.0, 0.0, 0.0 ), vr::vec3f( 0.0, 1.0, 0.0 ) );
 glMatrixMode( GL_MODELVIEW );
 glLoadMatrixf( s_viewMatrix.ptr() );
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
static void moveCamera( const vr::mat4f& transform )
{
 s_viewMatrix.product( s_viewMatrix, transform );
 glMatrixMode( GL_MODELVIEW );
 glLoadMatrixf( s_viewMatrix.ptr() );
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
void displayTextLine( const char* s, float x, float y )
{
 glRasterPos3d( x, y, 0.8 );

 for( unsigned int i = 0; s[i]; ++i )
  glutBitmapCharacter( GLUT_BITMAP_HELVETICA_12, s[i] );
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
void displayScreenMessage( const char* s, float x, float y )
{
 //glMatrixMode( GL_PROJECTION );
 //glPushMatrix();
 //glLoadIdentity();
 //glMatrixMode( GL_MODELVIEW );
 //glPushMatrix();
 //glLoadIdentity();
 //glDisable( GL_LIGHTING );

 glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT); 

 glDisable( GL_DEPTH_TEST );
 glColor3f( 0.0f, 0.0f, 0.0f );

 // Show FPS
 displayTextLine( s, x, y);

 //glEnable( GL_DEPTH_TEST );
 //glEnable( GL_LIGHTING );
 //glMatrixMode( GL_PROJECTION );
 //glPopMatrix();
 //glMatrixMode( GL_MODELVIEW );
 //glPopMatrix();

 glPopAttrib();
}


//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Update frames per second
static void updateFps()
{
 double elapsedTime = s_fpsTimer.elapsed();
 ++s_frameCounter;

 if( elapsedTime >= 1.0 )
 {
  double fps = s_frameCounter / elapsedTime;
  s_fpsMsg.format( "%.2f fps    %.2f ms", fps, ( elapsedTime * 1000.0 ) / s_frameCounter );

  s_frameCounter = 0;
  s_fpsTimer.restart();
 }
}

// Display callback
static void display()
{
 // Clear buffers
 glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

 // Set view matrix
 glMatrixMode(GL_MODELVIEW);
 glLoadIdentity();

 // Draw scene
 drawScene();

 // Update on-screen user information
 updateFps();
 displayScreenMessage( s_fpsMsg.toCharArray(), -19.95f, 19.90f );

 char msg[200];
 sprintf(msg, "Number of Points: %d", nPoints);
 displayScreenMessage( msg, -19.95f, 18.0f );

 glutSwapBuffers();
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Idle callback for performance measurements
static void idle()
{
 display();
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Reshape callback
static void reshape( int w, int h )
{
 static bool first = true;
 if( first )
 {
  glClearColor( 1.0f, 1.0f, 1.0f, 1.0f );
  glEnable( GL_DEPTH_TEST );

  glColorMaterial( GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE );
  glEnable( GL_COLOR_MATERIAL );

  glEnable( GL_NORMALIZE );

  // Create scene with valid OpenGL context
  createScene();

  // Disable V-Sync for benchmarking
  // Attention! Only works under Windows
  glewInit();
  wglSwapIntervalEXT( 0 );

  // Reset global timer
  s_fpsTimer.restart();

  first = false;
 }

 glViewport( 0, 0, w, h );
 s_screenWidth  = w;
 s_screenHeight = h;

 /* set projection */
 glMatrixMode(GL_PROJECTION);
 glLoadIdentity();
 gluOrtho2D(- maxValue - 3.0, maxValue + 3.0, - maxValue - 3.0, maxValue + 3.0);

 glMatrixMode(GL_MODELVIEW);	
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Keyboard callback
static void keyboard( unsigned char key, int x, int y )
{
 switch( key )
 {
  // Esc key: exit application
 case 27:
  exit( 0 );
  break;

 case '1': 
  nPoints = 5;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '2': 
  nPoints = 10;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '3': 
  nPoints = 30;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '4': 
  nPoints = 100;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '5': 
  nPoints = 500;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '6': 
  nPoints = 1000;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '7': 
  nPoints = 5000;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '8': 
  nPoints = 10000;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '9': 
  nPoints = 100000;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case '0': 
  nPoints = 5000000;
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case 'd':
  domain->reset();
  defaultScene();
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case 's':
  domain->square(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case 't':
  domain->triangle(nPoints, -maxValue, maxValue);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case 'c':
  domain->circle(nPoints, -maxValue, maxValue, 15.0);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case 'o':
  domain->on_circle(nPoints, -maxValue, maxValue, 20.0);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case 'r':
  domain->ring(nPoints, -maxValue, maxValue, 15.0, 20.0);
  convex_hull->reset(domain->get_points());
  delaunay->reset(domain->get_points());
  break;

 case 'j':
  convex_hull->jarvis();
  break;

 case 'v':
  convex_hull->scanning();
  break;

 case 'l':
  delaunay->divide_and_conquer();
  break;

 case 'b':
  if(delaunay->hasDelaunay())
   delaunay->hasDelaunay(false);
  else
   delaunay->hasDelaunay(true);
  break;

 case 'n':
  if(delaunay->hasVoronoi())
   delaunay->hasVoronoi(false);
  else
   delaunay->hasVoronoi(true);
  break;

 case 'm':
  if(domain->hasDomain())
   domain->hasDomain(false);
  else
   domain->hasDomain(true);
  break;

 case 'i':
  //convex_hull->reset(domain->get_points());
  if(convex_hull->hasElimination())
   convex_hull->hasElimination(false);
  else
   convex_hull->hasElimination(true);
  break;

 default:
  return;
 }

 glutPostRedisplay();
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Mouse callback
void mouse( int button, int state, int x, int y )
{
 if( button == GLUT_LEFT_BUTTON )
 {
  if( state == GLUT_DOWN )
  {
   s_prevMousePos.set( x, y );
   s_mouseDragActive = true;
  }
  else
  {
   s_mouseDragActive = false;
  }
 }
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
// Mouse drag callback
void drag( int x, int y )
{
 if( !s_mouseDragActive )
  return;

 vr::vec2f currentMousePos( x, y );
 vr::vec2f delta = currentMousePos - s_prevMousePos;
 s_prevMousePos = currentMousePos;

 vr::mat4f rotateX;
 rotateX.makeRotation( delta.x * s_lookSpeed, 0.0, 1.0, 0.0 );

 vr::mat4f rotateY;
 rotateY.makeRotation( delta.y * s_lookSpeed, 1.0, 0.0, 0.0 );

 rotateX.product( rotateX, rotateY );

 moveCamera( rotateX );

 glutPostRedisplay();
}

//------------------------------------------------------------------------
//------------------------------------------------------------------------
//////////////////////////////////////////////////////////////////////////
// Main entry point
//////////////////////////////////////////////////////////////////////////
#if 1
int main( int argc, char* argv[] )
{
 // Open GLUT
 glutInit( &argc, argv );
 glutInitDisplayMode( GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB );
 glutInitWindowSize( s_screenWidth, s_screenHeight );

#ifdef _DEBUG
 int windowId = glutCreateWindow( "Divide and Conquer Delaunay Triangulation" );
#else
 int windowId = glutCreateWindow( "Divide and Conquer Delaunay Triangulation" );
#endif

 glutReshapeFunc( reshape );
 glutDisplayFunc( display );
 glutIdleFunc( idle );
 glutKeyboardFunc( keyboard );
 glutMouseFunc( mouse );
 glutMotionFunc( drag );

 // Interact
 glutMainLoop();
 return 0;
}
#endif